﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using SilverMotion.EnvironmentTypes;
using SilverMotion.Motions;

// The application has the following scenario:
// We load the unique models needed for the game from a file, then generate the world by cloning models that are continuously appearing (markers, asteroids, gates, bullets).
// Whie moving forward, along the Z axis, objects remain behind the camera, which makes them obsolete. We recycle them by moving them ahead, to the current far end of the of our space highway, which ensures a perpetual never ending highway.
// We do a similar thing with asteroids, that have been hit by a bullet. Checking for collision for bullets and for ship bumping into an asteroid or a gate is done by the RayIntersection method, which provides exact information about any mesh, intersected by a 3D section defined by two points
namespace SpaceSimulator
{
    public partial class MainPage : UserControl
    {
        // Some game setting parameters
        const int verticalLanes = 4,                // number of vertical lanes in our space highway2
                  horizontalLanes = 1,              // number of horizontal lanes
                  gatesAhead = 10,                  // number of gates (along with the corresponging markers and asteroids) ahead of the ship
                  asteroidsBetween = 6,             // number of asteroids between two gates
                  particlesBetween = 20,            // number of particles between two gates
                  pointsPerGate = 50,               // number of points provided when passing through a gate
                  pointsPerAsteroid = 100,          // number of points provided when shot an asteroid
                  initialLifes = 3;                 // number of lifes at start

        const float laneSize = 140.0f,              // units
                    gatesDistance = 8000.0f,        // units
                    shipInitialSpeed = 600.0f,      // units per second
                    shipSpeedIncrement = 25.0f,     // units per second
                    gateSpeedDecrement = 50.0f,     // units per second of the ship's speed, that gets reduced, when it passes throug a gate
                    bulletSpeed = 1500.0f,          // units per second
                    bulletLife = 4000.0f,           // units
                    energyPerGate = 100.0f,         // the amount of energy a gate provides when passing throug it
                    energyPerShot = 20.0f,          // the amount of energy a shot costs 
                    energyPerGatesDistance = 40.0f, // the amount of energy cost for travelling the distance between two gates
                    fpsLimit = 30.0f;               // rendered frames per second 

        bool followingCamera = false;               // determines whether the camera should be strictly following the ship

        // Render branches used to store the models that we're going to use throughout the game and one for the gate that's closest to the ship
        RenderBranch ship, bullet, arrow, gate, mark, asteroid, particle, closestGate;
        // We'll use cameraVector and cameraTargetVector to store the camera's original position and target in relation to the ship. Ship rotation will hold the ship's rotation parameters, which we change when it moves between lanes. shipOldPos is the position of the ship in the previous frame. asteroidRotationAxis is the axis around which we'll rotate asteroids and it has to be furtherly normalized
        Vector3 cameraVector, cameraTargetVector, shipRotation, shipOldPos, asteroidRotationAxis = new Vector3(1, 1, 1);
        // horizontalLane and verticalLane are the coordinates of the lane that the ship is currently in or heading to. And "score" is the score.. same goes for lifes :)
        int horizontalLane, verticalLane, score, lifes = initialLifes;
        // Ship's speed. It is being constantly incremented and only passing through a gate can slow it down a bit
        float shipSpeed = shipInitialSpeed, shipEnergy = energyPerGate;
        // Used to enable/disable sounds and fps limit
        bool soundEnabled = true, fpsLimitEnabled = true;
        // Determines whether the game is running. Gets false when the ship "crashes" into an asteroid
        bool running = true;

        Random random = new Random();

        public class Bullet
        {
            public RenderBranch branch;
            public Vector3 oldPos;
            public float life;
        }
        List<Bullet> bullets = new List<Bullet>();

        public MainPage()
        {
            InitializeComponent();
        }

        private void environment_FileLoadCompleted(object sender, LoadCompletedEventArgs e)
        {
            // Set a background picture
            environment.SetBackgroundImage(environment.LoadedResources.FindTexture("Resources/space.zip/SPACE.JPG", false));

            environment.Update();

            // Get the render branches we will need to manipulate throughout the game
            ship = environment.Hierarchy.Find("Ship");
            arrow = environment.Hierarchy.Find("Arrow");
            bullet = environment.Hierarchy.Find("Bullet");
            gate = environment.Hierarchy.Find("Gate");
            mark = environment.Hierarchy.Find("Mark");
            asteroid = environment.Hierarchy.Find("Asteroid");
            particle = environment.Hierarchy.Find("Particle");

            // Remove from hierarchy the branches that we'll create dynamically in the world
            environment.Hierarchy.Remove(bullet);
            environment.Hierarchy.Remove(gate);
            environment.Hierarchy.Remove(mark);
            environment.Hierarchy.Remove(asteroid);
            environment.Hierarchy.Remove(particle);

            // Reset transformations to all the branches, because we'll be using our own transforms, not the predefined in modelling time
            ship.ResetTransformations();
            arrow.ResetTransformations();
            bullet.ResetTransformations();
            gate.ResetTransformations();
            mark.ResetTransformations();
            asteroid.ResetTransformations();
            particle.ResetTransformations();

            // Set the picking groups of objects which we don't want to interact with bullets or collide with ship to none. We leave only asteroids being able to get hit
            ship.PickingGroup = PickingGroups.None;
            arrow.PickingGroup = PickingGroups.None;
            bullet.PickingGroup = PickingGroups.None;
            gate.PickingGroup = PickingGroups.None;
            mark.PickingGroup = PickingGroups.None;
            particle.PickingGroup = PickingGroups.None;

            // Get camera position and target in relation to the ship, so we can move it along with the ship later
            cameraVector = environment.Camera.Position;
            cameraVector.Subtract(ship.TransformationReference.Position);
            cameraTargetVector = environment.Camera.Target;
            cameraTargetVector.Subtract(ship.TransformationReference.Position);

            // Set up the ship's starting position, normalize the asteroids' rotation axis, set some default value for the closest gate
            horizontalLane = verticalLanes / 2;
            verticalLane = horizontalLanes / 2;
            asteroidRotationAxis.Normalize();
            closestGate = gate;

            GenerateWorld();
            environment.Focus();

            progressBarEnergy.Visibility = Visibility.Visible;
            imageFpsLimit.Visibility = Visibility.Visible;
            imageSound.Visibility = Visibility.Visible;
            textBlockEnergy.Visibility = Visibility.Visible;
            textBlockLifes.Visibility = Visibility.Visible;
            textBlockScore.Visibility = Visibility.Visible;
            textBlockSpeed.Visibility = Visibility.Visible;

            UpdateInfo();

            // Start the motion, that we'll use for the game cycle
            environment.StartMotion(new CustomMotion(CustomMotionFunction));

            soundAmbient.Stop();
            soundAmbient.Play();
        }

        private void environment_KeyDown(object sender, KeyEventArgs e)
        {
            if (shipEnergy > 0)
            {
                if (e.Key == Key.Up)
                    verticalLane = Math.Max(verticalLane - 1, 0);

                if (e.Key == Key.Down)
                    verticalLane = Math.Min(verticalLane + 1, horizontalLanes - 1);

                if (e.Key == Key.Left)
                    horizontalLane = Math.Min(horizontalLane + 1, verticalLanes - 1);

                if (e.Key == Key.Right)
                    horizontalLane = Math.Max(horizontalLane - 1, 0);
            }
            else
            {
                soundNoAmmo.Stop();
                soundNoAmmo.Play();
            }

            if (e.Key == Key.Space)
                if (running)
                    AddBullet();
                else
                {
                    // Lost a life
                    imageAsteroidHit.Visibility = System.Windows.Visibility.Collapsed;
                    shipEnergy = energyPerGate;
                    if (lifes == 0)
                    {
                        // Game over, reset game
                        imageGameOver.Visibility = System.Windows.Visibility.Collapsed;
                        lifes = initialLifes;
                        shipEnergy = energyPerGate;
                        shipSpeed = shipInitialSpeed;
                        score = 0;
                        ship.TransformationReference.TranslateGlobal(0, 0, -gatesDistance);
                        shipOldPos.Z -= gatesDistance;

                        UpdateInfo();

                        soundAmbient.Play();
                    }
                    running = true;
                    environment.StartMotion(new CustomMotion(CustomMotionFunction));
                }
        }

        /// <summary>
        /// The function we will use for our main loop. It is a function used by the motion system and is started by the RenderEnvironment.StartMotion method
        /// </summary>
        /// <param name="time">The time in seconds, that has passed since the las rendered frame. We'll use this parameter to make any movements framerate independent</param>
        /// <returns>Returns true if the function should continue to be invoked and false otherwise</returns>
        private bool CustomMotionFunction(float time)
        {
            CheckShipCollisions();
            ProcessWorld(time);
            ProcessBullets(time);
            MoveShipAndCamera(time);
            SetArrow();

            return running;
        }

        /// <summary>
        /// Check if our ship has collided with something and check if it has gone through a gate
        /// </summary>
        private void CheckShipCollisions()
        {
            // Check if the ship has collided with something
            PickInfo info = null;
            if (environment.RayIntersection(shipOldPos, ship.TransformationReference.LocalPosition, ref info) != null)
            {
                soundExplode.Stop();
                soundExplode.Play();

                lifes--;
                UpdateInfo();

                imageAsteroidHit.Visibility = System.Windows.Visibility.Visible;
                if (lifes == 0)
                {
                    imageGameOver.Visibility = System.Windows.Visibility.Visible;
                    soundAmbient.Stop();
                }
                running = false; // crash it
            }

            // Check if the ship has passed a gate and then calculate the point at the moment of passing, which is where the ship path intersects with the gate's XY plane. Then check if the distance between this point and the gate is less than a gate's radius 
            // which would mean that the ship has successfully gone through the gate. We use this method here, because since the ship doesn't collide with polygons of a gate when it passes through it, we can't use the RayIntersection method.
            if ((shipOldPos.Z < closestGate.TransformationReference.LocalPosition.Z) && (ship.TransformationReference.LocalPosition.Z > closestGate.TransformationReference.LocalPosition.Z))
            {
                Vector3 through = shipOldPos;
                through.Lerp(ship.TransformationReference.LocalPosition, (closestGate.TransformationReference.LocalPosition.Z - shipOldPos.Z) / (ship.TransformationReference.LocalPosition.Z - shipOldPos.Z));
                if (through.Distance(closestGate.TransformationReference.LocalPosition) < 60)
                {
                    shipSpeed -= gateSpeedDecrement;
                    shipEnergy = energyPerGate;
                    score += pointsPerGate;

                    UpdateInfo();

                    soundGate.Stop();
                    soundGate.Play();
                }
            }
            shipOldPos = ship.TransformationReference.LocalPosition;
        }

        /// <summary>
        /// Check for obsolete objects in the world (ones that have remained behind the camera) and recycle them by moving them to the to the far end of the hightway. Rotate the asteroids and find the closest gate to the ship
        /// <param name="time">The time, passed by the CustomMotionFunction</param>
        /// </summary>
        private void ProcessWorld(float time)
        {
            for (int i = 0; i < environment.Hierarchy.Count; i++)
            {
                RenderBranch branch = environment.Hierarchy[i];

                // Translate any objects that are behind the camera to the far end of the current hightway, making it perperual. We translate glabally, because we want objects to move along the global Z axiz, not their local (in case they are rotated, for example)
                if (branch.TransformationReference.LocalPosition.Z < environment.Camera.Position.Z)
                    branch.TransformationReference.TranslateGlobal(0, 0, gatesAhead * gatesDistance);
                else
                {
                    // Make the asteroids slowly rotate around a local axis. The angle of rotation is based on the asteroid's Z position and "normalized" to values between -1 and 1 by the Sin function. This way all asteroids would have a different rotation speed and will appear to rotate randomly
                    if (branch.Name == asteroid.Name)
                        branch.TransformationReference.RotateLocalAxis(time * 2 * (float)Math.Sin(branch.TransformationReference.LocalPosition.Z), asteroidRotationAxis);

                    // Find the closest gate for further use
                    if (branch.Name == gate.Name)
                        if (((branch.TransformationReference.LocalPosition.Z < closestGate.TransformationReference.LocalPosition.Z) && (branch.TransformationReference.LocalPosition.Z > ship.TransformationReference.LocalPosition.Z)) || (closestGate.TransformationReference.LocalPosition.Z < ship.TransformationReference.LocalPosition.Z))
                            closestGate = branch;
                }
            }
        }

        /// <summary>
        /// Update game information included in the TextBlocks
        /// </summary>
        private void UpdateInfo()
        {
            textBlockLifes.Text = "Lifes: " + lifes.ToString();
            textBlockScore.Text = "Score: " + score.ToString();
            textBlockSpeed.Text = "Speed: " + shipSpeed.ToString("0.0");
        }

        /// <summary>
        /// Fire a new bullet from the ship. The bullet is oriented along with the ship, and because we move it along its local axis, it moves along its orientation.
        /// </summary>
        private void AddBullet()
        {
            if (shipEnergy < energyPerShot)
            {
                soundNoAmmo.Stop();
                soundNoAmmo.Play();
                return;
            }

            soundShoot.Stop();
            soundShoot.Play();

            shipEnergy -= energyPerShot;

            Bullet newBullet = new Bullet();
            newBullet.branch = bullet.Clone();
            newBullet.life = bulletLife;

            newBullet.branch.TransformationReference.Reset();
            newBullet.branch.TransformationReference.TranslateLocal(ship.TransformationReference.Position);
            newBullet.branch.TransformationReference.RotateLocalX(shipRotation.X);
            newBullet.branch.TransformationReference.RotateLocalZ(shipRotation.Z);

            environment.Hierarchy.Add(newBullet.branch);
            bullets.Add(newBullet);
        }

        /// <summary>
        /// Move bollets, check for obsolete ones and remove them, then check if the any of the rest has hit something and if so, take the respective actions
        /// </summary>
        /// <param name="time">The time, passed by the CustomMotionFunction</param>
        private void ProcessBullets(float time)
        {
            float bulletsVelocity = (shipSpeed + bulletSpeed) * time;
            for (int i = 0; i < bullets.Count; i++)
            {
                // Check for and remove obsolete bullets
                bullets[i].life -= bulletsVelocity;
                if (bullets[i].life <= 0)
                {
                    environment.Hierarchy.Remove(bullets[i].branch);
                    bullets.RemoveAt(i);
                    i--;
                }
                else
                {
                    // Move the bullet throug the scene space
                    bullets[i].branch.TransformationReference.TranslateLocal(0, 0, bulletsVelocity);

                    // Check if the bullet has hit something
                    PickInfo info = null;
                    if (environment.RayIntersection(bullets[i].oldPos, bullets[i].branch.TransformationReference.Position, ref info) != null)
                    {
                        RenderBranch victim = info.PickedBranch.ArchiParent;
                        if (victim.Name == asteroid.Name)
                        {
                            soundExplode.Stop();
                            soundExplode.Play();
                            victim.TransformationReference.TranslateGlobal(0, 0, gatesAhead * gatesDistance);
                            score += pointsPerAsteroid;
                            UpdateInfo();
                        }

                        environment.Hierarchy.Remove(bullets[i].branch);
                        bullets.RemoveAt(i);
                        i--;
                    }
                    else
                        bullets[i].oldPos = bullets[i].branch.TransformationReference.Position;
                }
            }
        }

        /// <summary>
        /// Move the ship towards the Z axis and set its position in the lanes, according to the horizontalLane and verticalLane parameters, which are set when the arrow keys are pressed.
        /// </summary>
        /// <param name="time">The time, passed by the CustomMotionFunction</param>
        private void MoveShipAndCamera(float time)
        {
            // The laneTarget vector is the location of the lane that the ship is headed. Based on it and lanePos, we calculate a vector pointing from the ship's position to the ship's target 
            // position and then we set it's length to the ship's velocity, so when we add it to lanePos, we get the new lane position of the ship, moved to the desired direction with the designated speed
            Vector3 laneTarget = new Vector3((horizontalLane + 0.5f) * laneSize, (verticalLane + 0.5f) * laneSize, 0);
            Vector3 lanePos = new Vector3(ship.TransformationReference.LocalPosition.X, ship.TransformationReference.LocalPosition.Y, 0);
            Vector3 targetVector = laneTarget;
            targetVector.Subtract(lanePos);
            targetVector.Length = Math.Min(time * shipSpeed * 0.3f, targetVector.Length);
            lanePos.Add(targetVector);

            // We calculate the shipRotation vector with Lerp in order to make the rotataion of the ship when it changes lane smoother. The vector we create for the Lerp is based on
            // the difference between the ship's current position and its target position as factors of how rotated it should be for both axes that correspond to pitch and yaw
            shipRotation.Lerp(new Vector3((float)(Math.Max(Math.Min((laneTarget.Y - lanePos.Y) / laneSize, 1.0), -1.0) * Math.PI / 24),
                                           0,
                                          (float)(Math.Max(Math.Min((laneTarget.X - lanePos.X) / laneSize, 1.0), -1.0) * Math.PI / 4)),
                              time * 10.0f);

            // Set the values we've calculated above
            ship.TransformationReference.Reset();
            ship.TransformationReference.TranslateLocal(lanePos.X, lanePos.Y, shipOldPos.Z + time * shipSpeed);
            ship.TransformationReference.RotateLocalX(shipRotation.X);
            ship.TransformationReference.RotateLocalZ(shipRotation.Z);

            // Now, setup the camera's position and target to go along with the ship
            Vector3 camPos;
            if (followingCamera)
            {
                camPos = cameraVector;
                camPos.Add(ship.TransformationReference.LocalPosition);
            }
            else
                camPos = new Vector3(verticalLanes * laneSize / 2, ship.TransformationReference.LocalPosition.Y + cameraVector.Y, ship.TransformationReference.LocalPosition.Z + cameraVector.Z);
            environment.Camera.Position = camPos;

            Vector3 camTarget = cameraTargetVector;
            camTarget.Add(ship.TransformationReference.LocalPosition);
            environment.Camera.Target = camTarget;

            shipSpeed += shipSpeedIncrement * time;
            shipEnergy -= (energyPerGatesDistance * shipSpeed * time) / gatesDistance;

            UpdateInfo();
            progressBarEnergy.Value = shipEnergy;
        }

        /// <summary>
        /// Setup the arrow which appears above the ship and points to the next gate
        /// </summary>
        private void SetArrow()
        {
            // Set the arrow's position in the scene. It will be 20 units above the ship's position
            Vector3 arrowPos = ship.TransformationReference.LocalPosition;
            arrowPos.Add(new Vector3(0, 20, 0));

            // Arrow direction is calculated by getting a vector that points from the ship's position to the gate's position and then calculating its spherical coordinates, which gives 
            // is the azimuth orientation as X, the elevation as Y, both of which we can directly apply to rotate the ship, and the distance as Z, which we can use to make the arrow's color
            Vector3 arrowDir = ship.TransformationReference.LocalPosition;
            arrowDir.Subtract(closestGate.TransformationReference.LocalPosition);
            arrowDir = arrowDir.GetSphericalCoords();

            // Apply the values we've calculated. We add Pi/2 when we rotate around X, because the azimuth angle we get starts from the X axis, and our ship is oriented to the Z axis, so 
            // to make this conversion from vector that points to X to one that point to Z, we simply have to add an angle of 90 degreece
            arrow.TransformationReference.Reset();
            arrow.TransformationReference.TranslateLocal(arrowPos);
            arrow.TransformationReference.RotateLocalY(arrowDir.X + (float)Math.PI / 2);
            arrow.TransformationReference.RotateLocalX(-arrowDir.Y);

            // Set the arrow's color according to the distance to the closest gate
            byte colorFactor = (byte)(0xff * Math.Min(arrowDir.Z, gatesDistance) / gatesDistance);
            arrow.MaterialReference.AmbientColor = Color.FromArgb(0xff, (byte)(0xff - colorFactor), 0x80, colorFactor);
        }

        /// <summary>
        /// Dynamically generate new objects ahead of the ship. This includes markers, gates and asteroids
        /// </summary>
        private void GenerateWorld()
        {
            // Generate one set of makers, asteroids and a gate at a time, "gatesAhead" times
            for (int gateIndex = 1; gateIndex <= gatesAhead; gateIndex++)
            {
                float generationZ = gateIndex * gatesDistance;

                // Generate a gate and start a keyframe animation motion for it
                RenderBranch newGate = gate.Clone();
                newGate.TransformationReference.LocalPosition = new Vector3(laneSize * (random.Next(verticalLanes) + 0.5f),
                                                                            laneSize * (random.Next(horizontalLanes) + 0.5f),
                                                                            generationZ);
                environment.Hierarchy.Add(newGate);
                environment.StartMotion(new KeyframeMotion(newGate, 0, 200, 30, true));

                // Generate the markers that appear where gates are
                for (int x = 0; x <= verticalLanes; x++)
                    for (int y = 0; y <= horizontalLanes; y++)
                    {
                        RenderBranch newMark = mark.Clone();
                        newMark.TransformationReference.LocalPosition = new Vector3(laneSize * x,
                                                                                    laneSize * y,
                                                                                    generationZ);
                        environment.Hierarchy.Add(newMark);
                    }

                // Generate asteroids between the gates and scale them by a random factor, so asteroids have different sizes
                for (int i = 0; i < asteroidsBetween; i++)
                {
                    RenderBranch newAsteroid = asteroid.Clone();
                    newAsteroid.TransformationReference.LocalPosition = new Vector3(laneSize * (random.Next(verticalLanes) + 0.5f),
                                                                                    laneSize * (random.Next(horizontalLanes) + 0.5f),
                                                                                    generationZ + gatesDistance * (i + 2) / (asteroidsBetween + 3));
                    newAsteroid.TransformationReference.ScaleLocal(1 + 2 * (float)random.NextDouble());
                    environment.Hierarchy.Add(newAsteroid);
                }

                // Generate particles between the gates
                for (int i = 0; i < particlesBetween; i++)
                {
                    RenderBranch newParticle = particle.Clone();
                    newParticle.TransformationReference.LocalPosition = new Vector3(laneSize * verticalLanes * (float)(random.NextDouble()),
                                                                                    laneSize * horizontalLanes * (float)(random.NextDouble()),
                                                                                    generationZ + gatesDistance * (i + 1) / (particlesBetween + 1));
                    environment.Hierarchy.Add(newParticle);
                }
            }
        }

        /// <summary>
        /// Loop on background music end
        /// </summary>
        private void soundAmbient_MediaEnded(object sender, RoutedEventArgs e)
        {
            // Loop the ambient music
            ((MediaElement)sender).Position = TimeSpan.Zero;
            ((MediaElement)sender).Play();
        }

        /// <summary>
        /// Turn on or off the frame speed limit
        /// </summary>
        private void imageFpsLimit_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            fpsLimitEnabled = !fpsLimitEnabled;
            if (fpsLimitEnabled)
            {
                environment.FPSLimit = fpsLimit;
                imageFpsLimit.Opacity = 1;
            }
            else
            {
                environment.FPSLimit = 1000.0f;
                imageFpsLimit.Opacity = 0.3;
            }
        }

        /// <summary>
        /// Turn on or off the sounds
        /// </summary>
        private void imageSound_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            soundEnabled = !soundEnabled;

            soundAmbient.IsMuted = !soundEnabled;
            soundExplode.IsMuted = !soundEnabled;
            soundGate.IsMuted = !soundEnabled;
            soundNoAmmo.IsMuted = !soundEnabled;
            soundShoot.IsMuted = !soundEnabled;
            if (soundEnabled)
            {
                imageSound.Opacity = 1;
            }
            else
            {
                imageSound.Opacity = 0.3;
            }
        }
    }
}
