﻿/// This is a sample project included in SilverMotion product contents.
/// Property of Postvision Ltd.
/// All rights reserved.
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using SilverMotion.EnvironmentTypes;
using SilverMotion.IO.FileImage;

namespace TerrainGenerator
{
    public partial class MainPage : UserControl
    {
        // Terrain vertex resolution
        const int terrainWidth = 48,
                  terrainHeight = 48;

        const int terrainVertexCount = terrainWidth * terrainHeight,
                  terrainPolyCount = (terrainWidth - 1) * (terrainHeight - 1) * 2;

        // Terrain mesh size
        const float terrainSizeX = 200.0f,
                    terrainSizeZ = 200.0f,
            // Brush settings
                    brushRadius = 0.1f,
                    brushRisePerSecond = 30.0f,
            // Terrain height limits
                    minHeight = -20.0f,
                    maxHeight = 50.0f;

        // Used to store the terraing branch for easy access throuthout the application
        RenderBranch terrainBranch;

        // Vertices array to hold vertex data
        Vertex[] vertices;

        // Various helper variables
        bool mbLeft, mbRight, modifying;
        float mx, my, dx, dy;
        float rx = 0.0f,
              ry = (float)Math.PI / 4;

        public MainPage()
        {
            InitializeComponent();

            // Create the RenderBranch for the terrain
            terrainBranch = new RenderBranch();
            terrainBranch.Name = "Terrain";

            // Create a TransformationResource for the terrain's RenderBranch
            terrainBranch.TransformationReference = new TransformationResource(environment, "Terrain Transform");

            // Create the MeshResource and fill it in with vertex and polygon data
            terrainBranch.MeshReference = new MeshResource(environment, "Terrain Mesh");
            terrainBranch.MeshReference.VertexCount = terrainVertexCount;
            terrainBranch.MeshReference.PolyCount = terrainPolyCount;

            // Generate vertices in the Vertex[] array and set them to the mesh
            vertices = new Vertex[terrainVertexCount];
            int index = 0;
            for (int ii = 0; ii < terrainHeight; ii++)
                for (int i = 0; i < terrainWidth; i++)
                {
                    vertices[index].Tu = 0.5f;
                    vertices[index].Tv = 0.5f;
                    vertices[index].Position = new Vector3((i / (float)(terrainWidth - 1)) * terrainSizeX - terrainSizeX / 2, 0.0f, terrainSizeZ / 2 - (ii / (float)(terrainHeight - 1)) * terrainSizeZ);
                    vertices[index].Normal = new Vector3(0.0f, 1.0f, 0.0f);
                    index++;
                }
            terrainBranch.MeshReference.SetVertices(0, 0, terrainVertexCount, vertices);

            // Generate polygons into the Triangle[] array and set them to the mesh. Polygons are described by 3 indices of the mesh's vertices
            Triangle[] triangles = new Triangle[terrainPolyCount];
            index = 0;
            for (int ii = 0; ii < terrainHeight - 1; ii++)
                for (int i = 0; i < terrainWidth - 1; i++)
                {
                    triangles[index].A = ii * terrainWidth + i;
                    triangles[index].B = triangles[index].A + 1;
                    triangles[index].C = triangles[index].A + terrainWidth;
                    index++;
                    triangles[index].A = ii * terrainWidth + i + 1;
                    triangles[index].B = triangles[index].A + terrainWidth;
                    triangles[index].C = triangles[index].B - 1;
                    index++;
                }
            terrainBranch.MeshReference.SetPolygons(0, 0, terrainPolyCount, triangles);

            // Load the texture we will use for coloring the terrain
            environment.LoadTexture(new FileImage(new Uri("gradient.png", UriKind.RelativeOrAbsolute)), false);

            // Create a MaterialResouce for the terrain and set its shaders and some of its properties
            terrainBranch.MaterialReference = new MaterialResource(environment, "Terrain Material");
            terrainBranch.MaterialReference.ClearShaders();
            terrainBranch.MaterialReference.AddShader(VertexShader.Lighting, ShaderColorOperation.Add);
            terrainBranch.MaterialReference.AddShader(TextureShader.ColorMap, ShaderColorOperation.Modulate, environment.LoadedResources.FindTexture("gradient.png", false));
            terrainBranch.MaterialReference.AmbientColor = Color.FromArgb(0xff, 0x80, 0x80, 0x80);
            terrainBranch.MaterialReference.Gloss = 20.0f;
            terrainBranch.MaterialReference.SpecularLevel = 0.4f;

            // Add the terrain's RenderBranch to the scene hierarchy
            environment.Hierarchy.Add(terrainBranch);

            // Create two dynamic lights to illuminate the scene
            LightResource light = new LightResource(environment, "Terrain Light1");
            light.Position = new Vector3(-terrainSizeX, 200, -terrainSizeZ);

            light = new LightResource(environment, "Terrain Light2");
            light.Position = new Vector3(terrainSizeX, 200, 0);

            // Setup camera's position an target
            environment.Camera.Position = new Vector3(0, 200, -terrainSizeZ);
            environment.Camera.Target = new Vector3(0, 0, 0);

            // Rotate terrain according to rx and ry
            RotateTerrain();

            // Update environment, so the set up scene is rendered
            environment.Update();
        }

        private void environment_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            mbLeft = true;
            StartSceneMotion(e);
        }

        private void environment_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            mbLeft = false;
        }

        private void environment_MouseRightButtonDown(object sender, MouseButtonEventArgs e)
        {
            e.Handled = true;
            mbRight = true;
            StartSceneMotion(e);
        }

        private void environment_MouseRightButtonUp(object sender, MouseButtonEventArgs e)
        {
            mbRight = false;
        }

        private void environment_MouseMove(object sender, MouseEventArgs e)
        {
            Point pos = e.GetPosition(environment);

            // Calculate mouse position delta
            dx = mx - (float)pos.X;
            dy = my - (float)pos.Y;

            // Rotate the terrain if mouse was clicked on an empty space before moved
            if (mbLeft & !modifying)
            {
                ry -= (float)(dx) / 100.0f;
                rx = (float)Math.Max(Math.Min(rx + (dy) / 100.0f, Math.PI / 6), -Math.PI / 6);
                RotateTerrain();
            }

            // Remember mouse position
            mx = (float)pos.X;
            my = (float)pos.Y;
        }

        /// <summary>
        /// The executive and condition function, used for the CustomMotion which we star when environment requires to be updated
        /// </summary>
        /// <param name="time">Time in seconds elapsed for the rendered frame</param>
        /// <returns>True if motion should continue, false if motion should terminate</returns>
        private bool CustomMotionFunction(float time)
        {
            // Modify terrain if mouse was clicked on the terrain surface
            if (modifying)
            {
                PickInfo pickInfo = new PickInfo();
                if (environment.Pick(mx, my, ref pickInfo) != null)
                    ModifyTerrain(pickInfo, time);
            }

            return mbLeft | mbRight;
        }

        /// <summary>
        /// Modifies the height and texture coordinates of the terrain according to the brush parameters and the provided PickInfo
        /// </summary>
        /// <param name="pickInfo">Information for the mouse intersecting the terrain</param>
        /// <param name="time">Time in seconds elapsed for the rendered frame. Used to modify the terrain height with a constant rate per time</param>
        private void ModifyTerrain(PickInfo pickInfo, float time)
        {
            float px = (0.5f + pickInfo.Position.X / terrainSizeX);
            float py = 1.0f - (0.5f + pickInfo.Position.Z / terrainSizeZ);
            int x1 = (int)Math.Max(Math.Min((px - brushRadius) * terrainWidth, terrainWidth), 0);
            int x2 = (int)Math.Max(Math.Min((px + brushRadius) * terrainWidth, terrainWidth), 0);
            int y1 = (int)Math.Max(Math.Min((py - brushRadius) * terrainHeight, terrainHeight), 0);
            int y2 = (int)Math.Max(Math.Min((py + brushRadius) * terrainHeight, terrainHeight), 0);

            for (int y = y1; y < y2; y++)
                for (int x = x1; x < x2; x++)
                {
                    float vu = px - (x / (float)terrainWidth);
                    float vv = py - (y / (float)terrainHeight);
                    float d = (float)Math.Sqrt(vu * vu + vv * vv);

                    if (d <= brushRadius)
                    {
                        d /= brushRadius;
                        int index = y * terrainWidth + x;
                        Vector3 pos = vertices[index].Position;
                        if (mbLeft)
                            pos.Y = (float)Math.Max(Math.Min(pos.Y + Math.Cos(d * Math.PI / 2) * brushRisePerSecond * time, maxHeight), minHeight);
                        if (mbRight)
                            pos.Y = (float)Math.Max(Math.Min(pos.Y - Math.Cos(d * Math.PI / 2) * brushRisePerSecond * time, maxHeight), minHeight);

                        if (pos.Y > 0.0f)
                            vertices[index].Tu = 0.5f + (pos.Y / maxHeight) * 0.49f;
                        else
                            vertices[index].Tu = 0.5f - (pos.Y / minHeight) * 0.49f;

                        vertices[index].Position = pos;
                    }
                }

            terrainBranch.MeshReference.SetVertices(0, 0, terrainVertexCount, vertices);
        }

        /// <summary>
        /// Rotates the terrain according to the ry and rx variables
        /// </summary>
        private void RotateTerrain()
        {
            terrainBranch.TransformationReference.Reset();
            terrainBranch.TransformationReference.RotateGlobalY(ry);
            terrainBranch.TransformationReference.RotateGlobalX(rx);
        }

        /// <summary>
        /// Starts the scene motion and checks if mouse was clicked on or out of the terrain surface.
        /// </summary>
        /// <param name="e">EventArgs for the mouse event</param>
        /// <remarks> Used by MouseDown events.</remarks>
        private void StartSceneMotion(MouseButtonEventArgs e)
        {
            Point pos = e.GetPosition(environment);
            PickInfo pickInfo = new PickInfo();
            modifying = environment.Pick((float)pos.X, (float)pos.Y, ref pickInfo) != null;

            environment.StartMotion(new CustomMotion(CustomMotionFunction));
        }
    }
}
