﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using SilverMotion.IO.File3DS;
using SilverMotion.IO.FileImage;
using SilverMotion.EnvironmentTypes;

namespace WebResources
{
    // Please note, that in order to be able to load resources from a web domain other than yours, you will need a crossdomain.xml file (included in the "Control" folder of SilverMotion commercial package) 
    // to be present in the root level of the domain where the resources are located.
    public partial class MainPage : UserControl
    {
        public enum Mode
        {
            Source, 
            Zip, 
            Textured3DS,
            Clean3DS
        }

        // This helper list will be used to add a files queued for loading and remove ones that have been already loaded, so when it's empty this would mean all queued files are loaded into the environment
        List<Uri> resourceQueue = new List<Uri>();
        
        // "mode" will help denoting which button has been pressed
        Mode mode = Mode.Source;

        // Uris to the resources that we'll be loading
        const string resourcesBase = "http://postvision.net/SilverMotion/Samples/";
        Uri zip = new Uri(resourcesBase + "Silver Frog.zip", UriKind.RelativeOrAbsolute);
        Uri textured3ds = new Uri(resourcesBase + "3ds/silverfrog.3ds", UriKind.RelativeOrAbsolute);
        Uri clean3ds = new Uri(resourcesBase + "3ds/silverfrog_clean.3ds", UriKind.RelativeOrAbsolute);
        Uri body_bm = new Uri(resourcesBase + "3ds/body_bm.jpg", UriKind.RelativeOrAbsolute);
        Uri body_cm = new Uri(resourcesBase + "3ds/body_cm.jpg", UriKind.RelativeOrAbsolute);
        Uri body_env = new Uri(resourcesBase + "3ds/body_env.jpg", UriKind.RelativeOrAbsolute);
        Uri eyes_cm = new Uri(resourcesBase + "3ds/eye_cm.jpg", UriKind.RelativeOrAbsolute);
        Uri eyes_env = new Uri(resourcesBase + "3ds/eye_env.jpg", UriKind.RelativeOrAbsolute);
        
        public MainPage()
        {
            InitializeComponent();
        }

        // Simply set the component's "Source" property to a file on the web.
        private void buttonSource_Click(object sender, RoutedEventArgs e)
        {
            mode = Mode.Source;

            environment.Source = null;
            environment.Source = zip.OriginalString;
        }

        // Queue a zip file, containing a 3DS model and its corresponging textures to be loaded asynchroniously.
        // Clearing the scene is not necessary. If omitted the loaded model and resources will simply be added to the scene.
        private void buttonImportZip_Click(object sender, RoutedEventArgs e)
        {
            mode = Mode.Zip;

            environment.Clear();
            environment.Import3DSFile(new File3DS(zip));
        }

        // Queue a 3DS file with textures to be loaded asynchroniously. All of the corresponding textures are on the server, along with the 
        // 3DS file and they will be automatically loaded.
        private void buttonImport3DS_Click(object sender, RoutedEventArgs e)
        {
            mode = Mode.Textured3DS;

            environment.Clear();
            environment.Import3DSFile(new File3DS(textured3ds));
        }

        // Queue a number of resources to be loaded asynchroniously. This time, the 3DS file we are loading was modelled without any textures, 
        // so we will load the textures separately. Because there are more than one resource here, we'll use the "resourceQueue" this time so
        // it can help us determine when all queued resources have been loaded when the "FileLoadCompleted" event appears.
        private void buttonImportCustom_Click(object sender, RoutedEventArgs e)
        {
            mode = Mode.Clean3DS;

            resourceQueue.Add(clean3ds);
            resourceQueue.Add(body_bm);
            resourceQueue.Add(body_cm);
            resourceQueue.Add(body_env);
            resourceQueue.Add(eyes_cm);
            resourceQueue.Add(eyes_env);

            environment.Clear();
            environment.Import3DSFile(new File3DS(clean3ds));
            environment.LoadTexture(new FileImage(body_bm), true);
            environment.LoadTexture(new FileImage(body_cm), false);
            environment.LoadTexture(new FileImage(body_env), false);
            environment.LoadTexture(new FileImage(eyes_cm), false);
            environment.LoadTexture(new FileImage(eyes_env), false);
        }

        // We get this event whenever a resource has been successfully loaded into the environment
        private void environment_FileLoadCompleted(object sender, SilverMotion.EnvironmentTypes.LoadCompletedEventArgs e)
        {
            resourceQueue.Remove(e.FileSource.UriSource);

            // Check if all queued resources are loaded
            if (resourceQueue.Count == 0)
            {
                // If we the textureless model was loaded, we need to put some materials and textures on it manually, so it will look like the others
                if (mode == Mode.Clean3DS)
                {
                    // We know all the resources are already loaded into the environment, so we get them
                    TextureResource tex_body_bm = environment.LoadedResources.FindTexture(body_bm.AbsoluteUri, true);
                    TextureResource tex_body_cm = environment.LoadedResources.FindTexture(body_cm.AbsoluteUri, false);
                    TextureResource tex_body_env = environment.LoadedResources.FindTexture(body_env.AbsoluteUri, false);
                    TextureResource tex_eyes_cm = environment.LoadedResources.FindTexture(eyes_cm.AbsoluteUri, false);
                    TextureResource tex_eyes_env = environment.LoadedResources.FindTexture(eyes_env.AbsoluteUri, false);

                    /// Get the mesh of the frog's body and set its material. The mesh has been named "SMFrog" when the model was created, 
                    /// and that's the name we'll search for in the scene hierarchy
                    #region Model Default
                    RenderBranch frog = environment.Hierarchy.Find("SMFrog");
                    frog.MaterialReference.SpecularColor = Color.FromArgb(0xff, 0xff, 0xff, 0xff);
                    frog.MaterialReference.Gloss = 40;
                    frog.MaterialReference.ClearShaders();
                    frog.MaterialReference.AddShader(TextureShader.NormalMap, ShaderColorOperation.None, tex_body_bm);
                    frog.MaterialReference.AddShader(PixelShader.Lighting, ShaderColorOperation.Add);
                    frog.MaterialReference.AddShader(TextureShader.ColorMap, ShaderColorOperation.Modulate, tex_body_cm);
                    frog.MaterialReference.AddShader(TextureShader.EnvMap, ShaderColorOperation.Add, tex_body_env);
                    #endregion

                    /// Following are a number of alternative shading blocks, producing various results. Shaders can be freely built by adding any of the shader operations on top of each other, giving the opportunity to create vast variety of shading effects
                    /// Uncomment any of the following regions to see the results of their shaing or perhaps try creating some shading of your own.
                    ///
                    /// By only adding a single shading operation, we can easily achieve reflective effect on the frog
                    #region Env Map
                    //frog.MaterialReference.ClearShaders();
                    //frog.MaterialReference.AddShader(TextureShader.EnvMap, ShaderColorOperation.Add, tex_body_env);
                    #endregion

                    /// In this block we add a "TextureShader.NormalMap" before the "TextureShader.EnvMap", resulting a bumped and reflective surface
                    #region Normal + Env Map
                    //frog.MaterialReference.ClearShaders();
                    //frog.MaterialReference.AddShader(TextureShader.NormalMap, ShaderColorOperation.None, tex_body_bm);
                    //frog.MaterialReference.AddShader(TextureShader.EnvMap, ShaderColorOperation.Add, tex_body_env);
                    #endregion

                    /// By having "VertexShader.Lighting" only, a dynamic lighting is produced. This lighing has many additional parameters, such as its material's DiffuseColor and SpecularColor, Gloss and SpecularLevel, as well as all the lights in the scene (LightResource) that all affect it simultaneously.
                    #region Dynamic Lighting
                    //frog.MaterialReference.DiffuseColor = Color.FromArgb(0xff, 0x40, 0xff, 0x00);
                    //frog.MaterialReference.SpecularColor = Color.FromArgb(0xff, 0xff, 0x00, 0xff);
                    //frog.MaterialReference.Gloss = 40;

                    //frog.MaterialReference.ClearShaders();
                    //frog.MaterialReference.AddShader(VertexShader.Lighting, ShaderColorOperation.Add);
                    #endregion

                    /// The "PixelShader.CellLighting" operation produces an effect which simulates cartoons. This shader has two specialized parameters - "CellDiffuseSegmentation" and "CellSpecularSegmentation". Additionally it's affected by the standard lighting properties and dynamic lights.
                    #region Dynamic Cell Shading
                    //environment.CellDiffuseSegmentation = 2;
                    //environment.CellSpecularSegmentation = 2;
                    //environment.AmbientColor = Color.FromArgb(0xff, 0xff, 0xff, 0xff);
                    //frog.MaterialReference.AmbientColor = Color.FromArgb(0xff, 0x10, 0x40, 0x00);
                    //frog.MaterialReference.DiffuseColor = Color.FromArgb(0xff, 0x40, 0xa0, 0x00);
                    //frog.MaterialReference.SpecularColor = Color.FromArgb(0xff, 0xd0, 0xd0, 0xd0);
                    //frog.MaterialReference.Gloss = 15;

                    //frog.MaterialReference.ClearShaders();
                    //frog.MaterialReference.AddShader(PixelShader.CellLighting, ShaderColorOperation.Add);
                    #endregion

                    // Get the mesh of the frog's eyes and set its material. This mesh has been named "Eyes" when the model was created.
                    RenderBranch eyes = environment.Hierarchy.Find("Eyes");
                    eyes.MaterialReference.SpecularColor = Color.FromArgb(0xff, 0xff, 0xff, 0xff);
                    eyes.MaterialReference.Gloss = 40;
                    eyes.MaterialReference.ClearShaders();
                    eyes.MaterialReference.AddShader(VertexShader.Lighting, ShaderColorOperation.Add);
                    eyes.MaterialReference.AddShader(TextureShader.ColorMap, ShaderColorOperation.Modulate, tex_eyes_cm);
                    eyes.MaterialReference.AddShader(TextureShader.EnvMap, ShaderColorOperation.Add, tex_eyes_env);
                }
            }
        }

        // Show a disasterous message if something bad happens
        private void environment_FileLoadFailed(object sender, LoadFailedEventArgs e)
        {
            MessageBox.Show("The resource " + e.FileSource.UriSource.AbsolutePath + " could not be loaded." + Environment.NewLine + e.Exception.Message, 
                            "Loading error", MessageBoxButton.OK);
        }
    }
}
